from typing import TypeVar

from cachetools import TTLCache, cached
from pydantic import BaseModel, Field

from .config import JWTConfig
from .enums import Algorithm
from .verify import extract_jwt_parts, verify_jwt, verify_temporal_claims

T = TypeVar("T", bound="BaseModel")


class JWT(BaseModel):
    token: str
    config: JWTConfig = Field(default_factory=JWTConfig)

    def __init__(
        self,
        *,
        token: str,
        config: JWTConfig | None = None,
        payload_class: type[T] | None = None,
    ):
        super().__init__(token=token, config=config)
        self._payload_class = payload_class

    def __hash__(self) -> int:
        return hash(self.token)

    @property
    @cached(cache=TTLCache(maxsize=1000, ttl=300))
    def _parts(self) -> tuple[dict, dict, bytes, bytes]:
        return extract_jwt_parts(self.token)

    @property
    def unverified_header(self) -> dict[str, str]:
        return self._parts[0]

    @property
    def header(self) -> dict[str, str]:
        if self.verify():
            return self.unverified_header
        raise ValueError("JWT is not valid")

    @property
    def algorithm(self) -> Algorithm:
        return Algorithm(self.unverified_header["alg"].upper())

    @property
    def unverified_payload(self) -> dict | T:
        if self._payload_class is not None:
            return self._payload_class.model_validate(self._parts[1])
        return self._parts[1]

    @property
    def payload(self) -> dict | T:
        if self.verify():
            return self.unverified_payload
        raise ValueError("JWT is not valid")

    @property
    def signature(self) -> bytes:
        return self._parts[2]

    @property
    def signing_input(self) -> bytes:
        return self._parts[3]

    def is_temporally_valid(self, *, raise_exception: bool = False) -> bool:
        try:
            if verify_temporal_claims(payload=self.unverified_payload):
                return True
            return False
        except Exception as e:
            if raise_exception:
                raise e
            return False

    def verify(self, expected_acr: str | list[str] | None = None) -> bool:
        return verify_jwt(
            token=self.token,
            jwk=self.config.key,
            expected_audience=self.config.audience,
            expected_issuer=self.config.issuer,
            expected_acr=expected_acr,
        )



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa

from ..utils import b64url_decode, b64url_encode
from .base import AbstractKey, KeyAlgorithm


class RSAAlgorithm(KeyAlgorithm):
    """RSA algorithm implementation
    (RS256, RS384, RS512, PS256, PS384, PS512)."""

    SUPPORTED_ALGORITHMS = {
        "RS256": hashes.SHA256,
        "RS384": hashes.SHA384,
        "RS512": hashes.SHA512,
        "PS256": hashes.SHA256,
        "PS384": hashes.SHA384,
        "PS512": hashes.SHA512,
    }

    @staticmethod
    def load_key(
        key: dict | bytes | rsa.RSAPrivateKey, password: bytes | None = None
    ) -> rsa.RSAPrivateKey:
        """
        Load RSA private key from JWK dict or PEM bytes.

        Args:
            key: Either a JWK dict or PEM-encoded private key bytes
            password: Optional password for encrypted PEM keys

        Returns:
            RSA private key object
        """
        if isinstance(key, rsa.RSAPrivateKey):
            return key
        if isinstance(key, dict):
            # Load from JWK
            n = int.from_bytes(b64url_decode(key["n"]), "big")
            e = int.from_bytes(b64url_decode(key["e"]), "big")
            d = int.from_bytes(b64url_decode(key["d"]), "big")
            return rsa.RSAPrivateNumbers(
                p=int.from_bytes(b64url_decode(key["p"]), "big"),
                q=int.from_bytes(b64url_decode(key["q"]), "big"),
                d=d,
                dmp1=int.from_bytes(b64url_decode(key["dp"]), "big"),
                dmq1=int.from_bytes(b64url_decode(key["dq"]), "big"),
                iqmp=int.from_bytes(b64url_decode(key["qi"]), "big"),
                public_numbers=rsa.RSAPublicNumbers(e, n),
            ).private_key(default_backend())

        # Load from DER
        return serialization.load_der_private_key(
            key, password=password, backend=default_backend()
        )

    @classmethod
    def sign(
        cls,
        *,
        data: bytes,
        key: dict | bytes | rsa.RSAPrivateKey,
        alg: str = "RS256",
        password: bytes | None = None,
    ) -> bytes:
        """
        Sign using RSA algorithms.

        Args:
            data: The data to sign
            key: Either a JWK dict or PEM-encoded private key bytes
            alg: The signing algorithm to use
                 (RS256, RS384, RS512, PS256, PS384, PS512)
            password: Optional password for encrypted PEM keys

        Returns:
            The signature
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported RSA algorithm: {alg}")

        privkey = cls.load_key(key, password)
        hash_alg = cls.SUPPORTED_ALGORITHMS[alg]

        if alg.startswith("RS"):
            return privkey.sign(data, padding.PKCS1v15(), hash_alg())
        else:  # PS
            return privkey.sign(
                data,
                padding.PSS(
                    mgf=padding.MGF1(hash_alg()),
                    salt_length=hashes.SHA256().digest_size,
                ),
                hash_alg(),
            )

    @classmethod
    def verify(
        cls,
        *,
        data: bytes,
        signature: bytes,
        key: dict | bytes,
        alg: str = "RS256",
    ) -> bool:
        """
        Verify RSA signature.

        Args:
            data: The data that was signed
            signature: The signature to verify
            key: Either a JWK dict or PEM-encoded public key bytes
            alg: The signing algorithm used
                 (RS256, RS384, RS512, PS256, PS384, PS512)
            password: Optional password for encrypted PEM keys

        Returns:
            True if signature is valid, False otherwise
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported RSA algorithm: {alg}")

        if isinstance(key, dict):
            # Load from JWK
            n = int.from_bytes(b64url_decode(key["n"]), "big")
            e = int.from_bytes(b64url_decode(key["e"]), "big")
            pubkey = rsa.RSAPublicNumbers(e, n).public_key(default_backend())
        else:
            # Load from DER
            pubkey = serialization.load_der_public_key(
                key,
                backend=default_backend(),
            )

        hash_alg = cls.SUPPORTED_ALGORITHMS[alg]

        try:
            if alg.startswith("RS"):
                pubkey.verify(signature, data, padding.PKCS1v15(), hash_alg())
            else:  # PS
                pubkey.verify(
                    signature,
                    data,
                    padding.PSS(
                        mgf=padding.MGF1(hash_alg()),
                        salt_length=hashes.SHA256().digest_size,
                    ),
                    hash_alg(),
                )
            return True
        except Exception:
            return False


class RSAKey(AbstractKey):
    """RSA key implementation."""

    def __init__(self, *, key: rsa.RSAPrivateKey, algorithm: str = "RS256"):
        self.key: rsa.RSAPrivateKey = key
        self.algorithm = algorithm

    @classmethod
    def generate(
        cls,
        *,
        algorithm: str = "RS256",
        key_size: int = 2048,
        public_exponent: int = 65537,
    ) -> "RSAKey":
        """Generate a new RSA key."""
        return RSAKey(
            key=rsa.generate_private_key(
                public_exponent=public_exponent,
                key_size=key_size,
                backend=default_backend(),
            ),
            algorithm=algorithm,
        )

    @classmethod
    def load_jwk(cls, key: dict) -> "RSAKey":
        """Load a key from JWK dict."""
        algorithm = key.get("alg", "RS256")
        return RSAKey(
            key=rsa.RSAPrivateNumbers(
                p=int.from_bytes(b64url_decode(key["p"]), "big"),
                q=int.from_bytes(b64url_decode(key["q"]), "big"),
                d=int.from_bytes(b64url_decode(key["d"]), "big"),
                dmp1=int.from_bytes(b64url_decode(key["dp"]), "big"),
                dmq1=int.from_bytes(b64url_decode(key["dq"]), "big"),
                iqmp=int.from_bytes(b64url_decode(key["qi"]), "big"),
                public_numbers=rsa.RSAPublicNumbers(
                    int.from_bytes(b64url_decode(key["e"]), "big"),
                    int.from_bytes(b64url_decode(key["n"]), "big"),
                ),
            ).private_key(default_backend()),
            algorithm=algorithm,
        )

    @classmethod
    def load_pem(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "RS256",
    ) -> "RSAKey":
        """Load a key from PEM."""
        key = super().load_pem(key, password)
        return RSAKey(key=key, algorithm=algorithm)

    @classmethod
    def load_der(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "RS256",
    ) -> "RSAKey":
        """Load a key from DER."""
        key = super().load_der(key, password)
        return RSAKey(key=key, algorithm=algorithm)

    def jwk(self) -> dict:
        """Get the JWK for the key."""
        public_key = self.key.public_key()
        return {
            "kty": "RSA",
            "alg": self.algorithm,
            "n": b64url_encode(
                public_key.public_numbers().n.to_bytes(
                    public_key.key_size // 8,
                    "big",
                )
            ),
            "e": b64url_encode(
                public_key.public_numbers().e.to_bytes(
                    public_key.key_size // 8,
                    "big",
                )
            ),
        }

    def public_key(self) -> rsa.RSAPublicKey:
        """Get the public key."""
        return self.key.public_key()

    @property
    def type(self) -> str:
        """Get the type of the key."""
        return "RSA"

    @property
    def key_size(self) -> int:
        """Get the size of the key."""
        return self.key.key_size

    def sign(self, data: bytes) -> bytes:
        """Sign data using the key."""
        return RSAAlgorithm.sign(data=data, key=self.key, alg=self.algorithm)

    def verify(self, data: bytes, signature: bytes) -> bool:
        """Verify signature using the key."""
        return RSAAlgorithm.verify(
            data=data,
            signature=signature,
            key=self.public_der(),
            alg=self.algorithm,
        )



import hashlib
from abc import ABC, abstractmethod

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from ..utils import b64url_encode


def convert_key_to_jwk(key: bytes) -> dict:
    """Convert PEM to dict."""
    # Check if the key is not in PEM format (doesn't start with BEGIN)
    if not key.startswith(b"-----BEGIN"):
        return {
            "kty": "oct",
            "k": b64url_encode(key),
        }

    public_key = serialization.load_pem_public_key(
        key, backend=default_backend()
    )
    if isinstance(public_key, RSAPublicKey):
        return {
            "kty": "RSA",
            "n": b64url_encode(
                public_key.public_numbers().n.to_bytes(256, "big")
            ),
            "e": b64url_encode(
                public_key.public_numbers().e.to_bytes(256, "big")
            ),
        }
    elif isinstance(public_key, EllipticCurvePublicKey):
        return {
            "kty": "EC",
            "crv": public_key.curve.name,
            "x": b64url_encode(
                public_key.public_numbers().x.to_bytes(256, "big")
            ),
            "y": b64url_encode(
                public_key.public_numbers().y.to_bytes(256, "big")
            ),
        }
    elif isinstance(public_key, Ed25519PublicKey):
        return {
            "kty": "OKP",
            "crv": "Ed25519",
            "x": b64url_encode(public_key.public_bytes_raw()),
        }
    else:
        raise ValueError("Unsupported algorithm")


class KeyAlgorithm(ABC):
    """Abstract base class for JWT algorithms."""

    @property
    @abstractmethod
    def SUPPORTED_ALGORITHMS(self) -> set[str]:
        """Set of supported algorithms for this implementation."""

    @staticmethod
    @abstractmethod
    def load_key(key: dict | bytes, password: bytes | None = None):
        """Load key from JWK dict or raw bytes."""

    @classmethod
    @abstractmethod
    def sign(
        cls,
        *,
        data: bytes,
        key,
        alg: str,
        password: bytes | None = None,
    ) -> bytes:
        """Sign data using the specified algorithm."""

    @classmethod
    @abstractmethod
    def verify(
        cls,
        *,
        data: bytes,
        signature: bytes,
        key: dict | bytes,
        alg: str,
        **kwargs,
    ) -> bool:
        """Verify signature using the specified algorithm."""


class AbstractKey(ABC):
    """Abstract base class for keys."""

    @classmethod
    @abstractmethod
    def generate(cls, **kwargs) -> "AbstractKey":
        """Generate a key."""

    @classmethod
    @abstractmethod
    def load_jwk(cls, key: dict) -> "AbstractKey":
        """Load a key from JWK dict."""

    @classmethod
    def load_pem(
        cls, key: bytes, password: bytes | None = None
    ) -> "AbstractKey":
        """Load a key from PEM."""
        return serialization.load_pem_private_key(
            key, password=password, backend=default_backend()
        )

    @classmethod
    def load_der(
        cls, key: bytes, password: bytes | None = None
    ) -> "AbstractKey":
        """Load a key from DER."""
        return serialization.load_der_private_key(
            key, password=password, backend=default_backend()
        )

    @classmethod
    def load(
        cls, key: dict | bytes, password: bytes | None = None
    ) -> "AbstractKey":
        """Load a key from JWK dict or PEM."""
        if isinstance(key, dict):
            return cls.load_jwk(key)
        if isinstance(key, bytes):
            return cls.load_der(key, password)

        raise ValueError("Invalid key data.")

    @abstractmethod
    def public_key(self):
        """Get the public key."""

    @abstractmethod
    def jwk(self) -> dict:
        """Get the JWK for the key."""

    def private_pem(self, password: bytes | None = None) -> bytes:
        """Get the private PEM for the key."""
        return self.key.private_bytes(
            serialization.Encoding.PEM,
            serialization.PrivateFormat.PKCS8,
            (
                serialization.NoEncryption()
                if password is None
                else serialization.BestAvailableEncryption(password)
            ),
        )

    def private_der(self, password: bytes | None = None) -> bytes:
        """Get the private PEM for the key."""
        return self.key.private_bytes(
            serialization.Encoding.DER,
            serialization.PrivateFormat.PKCS8,
            (
                serialization.NoEncryption()
                if password is None
                else serialization.BestAvailableEncryption(password)
            ),
        )

    def public_pem(self) -> bytes:
        """Get the public PEM for the key."""
        return self.key.public_key().public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    def public_der(self) -> bytes:
        """Get the public DER for the key."""
        return self.key.public_key().public_bytes(
            serialization.Encoding.DER,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    @property
    @abstractmethod
    def type(self) -> str:
        """Get the type of the key."""

    @property
    def kid(self) -> str:
        """Get the key ID for the key."""
        kid = hashlib.sha256(self.public_der()).hexdigest()
        return kid



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

from ..utils import b64url_decode, b64url_encode
from .base import AbstractKey, KeyAlgorithm


class EdDSAAlgorithm(KeyAlgorithm):
    """EdDSA algorithm implementation (Ed25519)."""

    SUPPORTED_ALGORITHMS = {"EdDSA", "EDDSA", "Ed25519", "ED25519"}

    @staticmethod
    def load_key(
        key: dict | bytes | ed25519.Ed25519PrivateKey,
        password: bytes | None = None,
    ) -> ed25519.Ed25519PrivateKey:
        """
        Load EdDSA private key from JWK dict or raw bytes.

        Args:
            key: Either a JWK dict or raw private key bytes
            password: Optional password for encrypted keys

        Returns:
            EdDSA private key object
        """
        if isinstance(key, ed25519.Ed25519PrivateKey):
            return key
        if isinstance(key, dict):
            return ed25519.Ed25519PrivateKey.from_private_bytes(
                b64url_decode(key["d"])
            )
        return serialization.load_der_private_key(
            key, password=password, backend=default_backend()
        )

    @classmethod
    def sign(
        cls,
        *,
        data: bytes,
        key: dict | bytes | ed25519.Ed25519PrivateKey,
        alg: str = "EdDSA",
        password: bytes | None = None,
    ) -> bytes:
        """
        Sign using EdDSA algorithm.

        Args:
            data: The data to sign
            key: Either a JWK dict or raw private key bytes
            alg: The signing algorithm to use (must be "EdDSA")
            password: Optional password for encrypted keys

        Returns:
            The signature
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported EdDSA algorithm: {alg}")

        privkey = cls.load_key(key, password)
        return privkey.sign(data)

    @classmethod
    def verify(
        cls,
        *,
        data: bytes,
        signature: bytes,
        key: dict | bytes | ed25519.Ed25519PublicKey,
        alg: str = "EdDSA",
    ) -> bool:
        """
        Verify EdDSA signature.

        Args:
            data: The data that was signed
            signature: The signature to verify
            key: Either a JWK dict or raw public key bytes
            alg: The signing algorithm used (must be "EdDSA")
            password: Optional password for encrypted keys

        Returns:
            True if signature is valid, False otherwise
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported EdDSA algorithm: {alg}")

        if isinstance(key, dict):
            pubkey = ed25519.Ed25519PublicKey.from_public_bytes(
                b64url_decode(key["x"])
            )
        else:
            pubkey = serialization.load_der_public_key(
                key, backend=default_backend()
            )

        try:
            pubkey.verify(signature, data)
            return True
        except Exception:
            return False


class EdDSAKey(AbstractKey):
    """EdDSA key implementation."""

    def __init__(
        self, *, key: ed25519.Ed25519PrivateKey, algorithm: str = "EdDSA"
    ):
        self.key = key
        self.algorithm = algorithm

    @classmethod
    def generate(
        cls,
        *,
        algorithm: str = "EdDSA",
    ) -> "EdDSAKey":
        """Generate a new EdDSA key."""
        return EdDSAKey(
            key=ed25519.Ed25519PrivateKey.generate(), algorithm=algorithm
        )

    @classmethod
    def load_jwk(cls, key: dict) -> "EdDSAKey":
        """Load a key from JWK dict."""
        algorithm = key.get("alg", "EdDSA")
        return EdDSAKey(
            key=ed25519.Ed25519PrivateKey.from_private_bytes(
                b64url_decode(key["d"])
            ),
            algorithm=algorithm,
        )

    @classmethod
    def load_pem(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "EdDSA",
    ) -> "EdDSAKey":
        """Load a key from PEM."""
        key = super().load_pem(key, password)
        return EdDSAKey(key=key, algorithm=algorithm)

    @classmethod
    def load_der(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "EdDSA",
    ) -> "EdDSAKey":
        """Load a key from DER."""
        key = super().load_der(key, password)
        return EdDSAKey(key=key, algorithm=algorithm)

    def jwk(self) -> dict:
        """Get the JWK for the key."""
        public_key = self.key.public_key()
        return {
            "kty": "OKP",
            "crv": "Ed25519",
            "x": b64url_encode(public_key.public_bytes_raw()),
            "alg": "EdDSA",
            "use": "sig",
        }

    def public_key(self) -> ed25519.Ed25519PublicKey:
        """Get the public key."""
        return self.key.public_key()

    @property
    def type(self) -> str:
        """Get the type of the key."""
        return "EdDSA"

    def sign(self, data: bytes) -> bytes:
        """Sign data using the key."""
        return EdDSAAlgorithm.sign(data=data, key=self.key, alg=self.algorithm)

    def verify(self, data: bytes, signature: bytes) -> bool:
        """Verify signature using the key."""
        return EdDSAAlgorithm.verify(
            data=data,
            signature=signature,
            key=self.public_der(),
            alg=self.algorithm,
        )



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, utils

from ..utils import b64url_decode, b64url_encode
from .base import AbstractKey, KeyAlgorithm


class ECDSAAlgorithm(KeyAlgorithm):
    """ECDSA algorithm implementation (ES256, ES384, ES512)."""

    SUPPORTED_ALGORITHMS = {
        "ES256": ec.SECP256R1,
        "ES384": ec.SECP384R1,
        "ES512": ec.SECP521R1,
    }

    @classmethod
    def load_key(
        cls,
        key: dict | bytes | ec.EllipticCurvePrivateKey,
        alg: str,
        password: bytes | None = None,
    ) -> ec.EllipticCurvePrivateKey:
        """
        Load ECDSA private key from JWK dict or PEM bytes.

        Args:
            key: Either a JWK dict or PEM-encoded private key bytes
            alg: The signing algorithm (ES256, ES384, ES512)
            password: Optional password for encrypted PEM keys

        Returns:
            ECDSA private key object
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported ECDSA algorithm: {alg}")

        if isinstance(key, ec.EllipticCurvePrivateKey):
            return key
        if isinstance(key, dict):
            # Load from JWK
            curve = cls.SUPPORTED_ALGORITHMS[alg]()
            x = int.from_bytes(b64url_decode(key["x"]), "big")
            y = int.from_bytes(b64url_decode(key["y"]), "big")
            d = int.from_bytes(b64url_decode(key["d"]), "big")

            return ec.EllipticCurvePrivateNumbers(
                d, ec.EllipticCurvePublicNumbers(x, y, curve)
            ).private_key(default_backend())
        # Load from DER
        return serialization.load_der_private_key(
            key, password=password, backend=default_backend()
        )

    @classmethod
    def sign(
        cls,
        *,
        data: bytes,
        key: dict | bytes | ec.EllipticCurvePrivateKey,
        alg: str = "ES256",
        password: bytes | None = None,
    ) -> bytes:
        """
        Sign using ECDSA algorithms.

        Args:
            signing_input: The data to sign
            key: Either a JWK dict or PEM-encoded private key bytes
            alg: The signing algorithm to use (ES256, ES384, ES512)
            password: Optional password for encrypted PEM keys

        Returns:
            The signature
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported ECDSA algorithm: {alg}")

        privkey = cls.load_key(key, alg, password)

        # Sign and format signature
        signature = privkey.sign(data, ec.ECDSA(hashes.SHA256()))
        r, s = utils.decode_dss_signature(signature)
        size = (privkey.curve.key_size + 7) // 8
        return r.to_bytes(size, "big") + s.to_bytes(size, "big")

    @classmethod
    def verify(
        cls,
        *,
        data: bytes,
        signature: bytes,
        key: dict | bytes,
        alg: str = "ES256",
    ) -> bool:
        """
        Verify ECDSA signature.

        Args:
            data: The data that was signed
            signature: The signature to verify
            key: Either a JWK dict or PEM-encoded public key bytes
            alg: The signing algorithm used (ES256, ES384, ES512)
            password: Optional password for encrypted PEM keys

        Returns:
            True if signature is valid, False otherwise
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported ECDSA algorithm: {alg}")

        if isinstance(key, dict):
            # Load from JWK
            curve = cls.SUPPORTED_ALGORITHMS[alg]()
            x = int.from_bytes(b64url_decode(key["x"]), "big")
            y = int.from_bytes(b64url_decode(key["y"]), "big")
            pubkey = ec.EllipticCurvePublicNumbers(x, y, curve).public_key(
                default_backend()
            )
        else:
            # Load from DER
            pubkey = serialization.load_der_public_key(
                key, backend=default_backend()
            )

        try:
            # Reconstruct signature from r and s components
            size = (pubkey.curve.key_size + 7) // 8
            r = int.from_bytes(signature[:size], "big")
            s = int.from_bytes(signature[size:], "big")
            sig = utils.encode_dss_signature(r, s)

            pubkey.verify(sig, data, ec.ECDSA(hashes.SHA256()))
            return True
        except Exception:
            return False


class ECDSAKey(AbstractKey):
    """ECDSA key implementation."""

    def __init__(
        self, *, key: ec.EllipticCurvePrivateKey, algorithm: str = "ES256"
    ):
        self.key = key
        self.algorithm = algorithm

    @classmethod
    def generate(
        cls,
        *,
        algorithm: str = "ES256",
    ) -> "ECDSAKey":
        """Generate a new ECDSA key."""
        if algorithm == "ES256":
            curve = ec.SECP256R1()
        elif algorithm == "ES384":
            curve = ec.SECP384R1()
        elif algorithm == "ES512":
            curve = ec.SECP521R1()
        else:
            raise ValueError(f"Unsupported ECDSA algorithm: {algorithm}")

        return ECDSAKey(
            key=ec.generate_private_key(curve, default_backend()),
            algorithm=algorithm,
        )

    @classmethod
    def load_jwk(cls, key: dict) -> "ECDSAKey":
        """Load a key from JWK dict."""
        algorithm = key.get("alg", "ES256")
        return ECDSAKey(
            key=ec.EllipticCurvePrivateNumbers(
                int.from_bytes(b64url_decode(key["d"]), "big"),
                ec.EllipticCurvePublicNumbers(
                    int.from_bytes(b64url_decode(key["x"]), "big"),
                    int.from_bytes(b64url_decode(key["y"]), "big"),
                    ec.SECP256R1(),
                ),
            ).private_key(default_backend()),
            algorithm=algorithm,
        )

    @classmethod
    def load_pem(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "ES256",
    ) -> "ECDSAKey":
        """Load a key from PEM."""
        key = super().load_pem(key, password)
        return ECDSAKey(key=key, algorithm=algorithm)

    @classmethod
    def load_der(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "ES256",
    ) -> "ECDSAKey":
        """Load a key from DER."""
        key = super().load_der(key, password)
        return ECDSAKey(key=key, algorithm=algorithm)

    def jwk(self) -> dict:
        """Get the JWK for the key."""
        public_key = self.key.public_key()
        CURVE_NAME_TO_JWK_CRV = {
            "secp256r1": "P-256",
            "secp384r1": "P-384",
            "secp521r1": "P-521",
        }
        return {
            "kty": "EC",
            "alg": self.algorithm,
            "crv": CURVE_NAME_TO_JWK_CRV[public_key.curve.name],
            "x": b64url_encode(
                public_key.public_numbers().x.to_bytes(
                    public_key.curve.key_size // 8, "big"
                )
            ),
            "y": b64url_encode(
                public_key.public_numbers().y.to_bytes(
                    public_key.curve.key_size // 8, "big"
                )
            ),
        }

    def public_key(self) -> ec.EllipticCurvePublicKey:
        """Get the public key."""
        return self.key.public_key()

    @property
    def type(self) -> str:
        """Get the type of the key."""
        return "ECDSA"

    def sign(self, data: bytes) -> bytes:
        """Sign data using the key."""
        return ECDSAAlgorithm.sign(data=data, key=self.key, alg=self.algorithm)

    def verify(self, data: bytes, signature: bytes) -> bool:
        """Verify signature using the key."""
        return ECDSAAlgorithm.verify(
            data=data,
            signature=signature,
            key=self.public_der(),
            alg=self.algorithm,
        )



import httpx
import pytest

from src.usso_jwt import exceptions, sign, verify
from src.usso_jwt.algorithms import AbstractKey


def test_fetch_jwk():
    jwks_url = "https://www.googleapis.com/oauth2/v3/certs"
    jwks = httpx.get(jwks_url).json()
    jwk = jwks["keys"][0]
    fetched_jwk = verify.fetch_jwk(jwks_url=jwks_url, kid=jwk["kid"])
    assert fetched_jwk is not None
    assert fetched_jwk["kid"] == jwk["kid"]
    assert fetched_jwk["alg"] == jwk["alg"]
    assert fetched_jwk["kty"] == jwk["kty"]
    assert fetched_jwk["n"] is not None
    assert fetched_jwk["e"] is not None


def test_fetch_failed_jwk():
    with pytest.raises(exceptions.JWKNotFoundError):
        verify.fetch_jwk(
            jwks_url="https://www.googleapis.com/oauth2/v3/certs", kid="123"
        )


def test_invalid_signature(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    jwt = jwt[:-1]  # + chr(ord(jwt[-1]) + 1)
    with pytest.raises(exceptions.JWTInvalidFormatError):
        verify.verify_jwt(token=jwt, jwk=test_key.jwk())


def test_expired_payload(
    test_expired_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_expired_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTExpiredError):
        verify.verify_jwt(token=jwt, jwk=test_key.jwk())


def test_nbf_future_payload(
    test_future_nbf_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_future_nbf_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTNotValidYetError):
        verify.verify_jwt(token=jwt, jwk=test_key.jwk())


def test_future_payload(
    test_future_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_future_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTIssuedInFutureError):
        verify.verify_jwt(token=jwt, jwk=test_key.jwk())


def test_missing_audience(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    payload = test_valid_payload.copy()
    payload.pop("aud", None)
    jwt = sign.generate_jwt(
        header=test_header,
        payload=payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTMissingAudienceError):
        verify.verify_jwt(
            token=jwt, jwk=test_key.jwk(), expected_audience="test_jwt"
        )


def test_invalid_audience(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTInvalidAudienceError):
        verify.verify_jwt(
            token=jwt, jwk=test_key.jwk(), expected_audience="test_jwt"
        )


def test_invalid_acr_audience(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTInvalidACRError):
        verify.verify_jwt(
            token=jwt,
            jwk=test_key.jwk(),
            expected_acr="refresh",
        )


def test_invalid_issuer(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTInvalidIssuerError):
        verify.verify_jwt(
            token=jwt,
            jwk=test_key.jwk(),
            expected_issuer="test_jwt",
        )



import json
import time

import httpx
from cachetools import TTLCache, cached

from .algorithms import get_algorithm
from .exceptions import (
    JWKNotFoundError,
    JWTExpiredError,
    JWTInvalidACRError,
    JWTInvalidAudienceError,
    JWTInvalidFormatError,
    JWTInvalidIssuerError,
    JWTInvalidSignatureError,
    JWTIssuedInFutureError,
    JWTMissingAudienceError,
    JWTNotValidYetError,
)
from .utils import b64url_decode, b64url_encode


def extract_jwt_parts(token: str) -> tuple[dict, dict, bytes, bytes]:
    try:
        header_b64, payload_b64, signature_b64 = token.split(".")
        header = json.loads(b64url_decode(header_b64))
        payload = json.loads(b64url_decode(payload_b64))
        signature = b64url_decode(signature_b64)
        signing_input = f"{header_b64}.{payload_b64}".encode()
        return header, payload, signature, signing_input
    except (ValueError, json.JSONDecodeError) as e:
        raise JWTInvalidFormatError() from e


@cached(cache=TTLCache(maxsize=1000, ttl=3600))
def fetch_jwk(*, jwks_url: str, kid: str) -> dict | None:
    jwks = httpx.get(jwks_url).json()
    for key in jwks["keys"]:
        if key["kid"] == kid:
            return key
    raise JWKNotFoundError()


def verify_temporal_claims(*, payload: dict):
    now = int(time.time())
    if "exp" in payload and now >= payload["exp"]:
        raise JWTExpiredError()
    if "nbf" in payload and now < payload["nbf"]:
        raise JWTNotValidYetError()
    if "iat" in payload and now < payload["iat"] - 60:
        raise JWTIssuedInFutureError()
    return True


def verify_claims(
    *,
    payload: dict,
    expected_audience: str | list[str] | None = None,
    expected_acr: str | list[str] | None = None,
    expected_issuer: str | list[str] | None = None,
) -> bool:
    """
    Verify additional JWT claims like audience, acr, and issuer if they
    are present.

    Args:
        payload: The JWT payload
        expected_audience: The expected audience value(s) to validate against.
                           Can be a single string or list of strings.
                           If provided, the token MUST have an aud claim
                           that matches one of the expected values.
        expected_acr:      The expected acr value(s) to validate against.
                           Can be a single string or list of strings.
        expected_issuer:   The expected issuer value(s) to validate against.
                           Can be a single string or list of strings.

    Returns:
        bool: True if all claims are valid

    Raises:
        JWTInvalidAudienceError: If the audience claim is invalid
        JWTMissingAudienceError: If the audience claim is missing
                                 and expected_audience is provided
        JWTInvalidACRError: If the acr claim is invalid
        JWTInvalidIssuerError: If the issuer claim is invalid
    """
    # Handle audience validation
    if expected_audience is not None:
        if "aud" not in payload:
            raise JWTMissingAudienceError()
        if isinstance(expected_audience, str):
            expected_audience = [expected_audience]
        if payload["aud"] not in expected_audience:
            raise JWTInvalidAudienceError()

    # Handle ACR validation
    if "acr" in payload and expected_acr is not None:
        if isinstance(expected_acr, str):
            expected_acr = [expected_acr]
        if payload["acr"] not in expected_acr:
            raise JWTInvalidACRError()

    # Handle issuer validation
    if "iss" in payload and expected_issuer is not None:
        if isinstance(expected_issuer, str):
            expected_issuer = [expected_issuer]
        if payload["iss"] not in expected_issuer:
            raise JWTInvalidIssuerError()

    return True


def verify_signature(
    *, alg: str, key: dict, data: bytes | str | dict, signature: bytes
) -> bool:
    """
    Verify a JWT signature using the specified algorithm and key.

    Args:
        alg: The algorithm used for signing
        key: The verification key (JWK dict)
        signing_input: The data that was signed
        signature: The signature to verify

    Returns:
        bool: True if the signature is valid, False otherwise

    Raises:
        JWTInvalidSignatureError: If the signature verification fails
    """
    if isinstance(data, dict):
        signing_input = b64url_encode(json.dumps(data)).encode()
    elif isinstance(data, str):
        signing_input = data.encode()
    else:
        signing_input = data

    algorithm = get_algorithm(alg)
    if not algorithm.verify(
        data=signing_input, signature=signature, key=key, alg=alg
    ):
        raise JWTInvalidSignatureError()
    return True


def verify_jwt(
    *,
    token: str,
    jwks_url: str = None,
    kid: str = None,
    jwk: dict | None = None,
    expected_audience: str | list[str] | None = None,
    expected_acr: str | list[str] | None = None,
    expected_issuer: str | list[str] | None = None,
) -> bool:
    header, payload, signature, signing_input = extract_jwt_parts(token)
    if jwk is None:
        jwk = fetch_jwk(jwks_url=jwks_url, kid=kid)

    if not verify_signature(
        alg=header["alg"], key=jwk, data=signing_input, signature=signature
    ):
        raise JWTInvalidSignatureError()

    verify_temporal_claims(payload=payload)
    verify_claims(
        payload=payload,
        expected_audience=expected_audience,
        expected_acr=expected_acr,
        expected_issuer=expected_issuer,
    )

    return True
